import { parser } from 'lezer-cql';
import { syntaxTree, LRLanguage, LanguageSupport } from '@codemirror/language';
import { ifIn, completeFromList, insertCompletionText } from '@codemirror/autocomplete';
import { linter } from '@codemirror/lint';

function reverseFind(text, char) {
    let i = text.length;
    while (i > 0) {
        if (text[i] === char) {
            return i;
        }
        i--;
    }
    return -1;
}
const cqlDateInstanceCompletionSource = () => {
    return ifIn(["DateInstantString"], (context) => {
        const dateStrStart = reverseFind(context.state.sliceDoc(0, context.pos), "'") + 1;
        const curStr = context.state.sliceDoc(dateStrStart, context.pos);
        const offset = context.pos - dateStrStart;
        const quoteEnd = context.state.sliceDoc(context.pos).indexOf("'");
        if (quoteEnd === -1) {
            return null;
        }
        const dateStrEnd = context.pos + quoteEnd;
        const completionTemplate = "1000-01-01";
        let completionText = curStr + completionTemplate.substring(offset);
        const options = [];
        if (dateStrEnd !== context.pos) {
            options.push({
                label: context.state.sliceDoc(dateStrStart, dateStrEnd), type: "constant",
            });
        }
        options.push({
            label: completionText, type: "constant",
        });
        const completionResult = ({
            from: dateStrStart, to: dateStrEnd, options
        });
        return completionResult;
    });
};
const cqlTimestampInstanceCompletionSource = () => {
    return ifIn(["TimestampInstantString"], (context) => {
        const dateStrStart = reverseFind(context.state.sliceDoc(0, context.pos), "'") + 1;
        const curStr = context.state.sliceDoc(dateStrStart, context.pos);
        const offset = context.pos - dateStrStart;
        const quoteEnd = context.state.sliceDoc(context.pos).indexOf("'");
        if (quoteEnd === -1) {
            return null;
        }
        const dateStrEnd = context.pos + quoteEnd;
        const completionTemplate = "1000-01-01T00:00:00Z";
        let completionText = curStr + completionTemplate.substring(offset);
        const completionEndTemplate = "1000-01-01T23:59:59Z";
        let completionEndText = curStr + completionEndTemplate.substring(offset);
        const options = [];
        if (dateStrEnd !== context.pos) {
            options.push({
                label: context.state.sliceDoc(dateStrStart, dateStrEnd), type: "constant",
            });
        }
        options.push({ label: completionText, type: "constant" });
        options.push({ label: completionEndText, type: "constant" });
        const completionResult = ({
            from: dateStrStart, to: dateStrEnd, options
        });
        return completionResult;
    });
};
const cqlKeywordCompletionSource = () => {
    return ifIn(["BinaryComparisonPredicate", "BooleanExpression"], completeFromList([
        { displayLabel: "and", label: "and ", type: "keyword" },
        { displayLabel: "or", label: "or ", type: "keyword" },
        { displayLabel: "not", label: "not ", type: "keyword" },
        { displayLabel: "like", label: "like ", type: "keyword" },
        { displayLabel: "between", label: "between ", type: "keyword" },
        { displayLabel: "is", label: "is ", type: "keyword" },
        { displayLabel: "in", label: "in ", type: "keyword" },
    ]));
};
const cqlConstantCompletionSource = () => {
    return ifIn(["BinaryComparisonPredicate", "BooleanExpression"], completeFromList([
        { label: "null", type: "constant" },
        { label: "EMPTY", type: "constant" },
        { label: "true", type: "constant" },
        { label: "false", type: "constant" },
    ]));
};
function applyFunctionCompletion(view, completion, from, to) {
    const text = completion.label + "()";
    const pos = from + text.length - 1;
    view.dispatch(insertCompletionText(view.state, text, from, to));
    view.dispatch({
        selection: { anchor: pos, head: pos },
    });
}
function createFunctionCompetion(label) {
    return {
        label,
        type: "function",
        apply: applyFunctionCompletion,
    };
}
const functionCompletionList = /*@__PURE__*/[
    "T_AFTER",
    "T_BEFORE",
    "T_CONTAINS",
    "T_DISJOINT",
    "T_DURING",
    "T_EQUALS",
    "T_FINISHEDBY",
    "T_FINISHES",
    "T_INTERSECTS",
    "T_MEETS",
    "T_METBY",
    "T_OVERLAPPEDBY",
    "T_OVERLAPS",
    "T_STARTEDBY",
    "T_STARTS",
    "S_INTERSECTS",
    "S_EQUALS",
    "S_DISJOINT",
    "S_TOUCHES",
    "S_WITHIN",
    "S_OVERLAPS",
    "S_CROSSES",
    "S_CONTAINS",
    "A_EQUALS",
    "A_CONTAINS",
    "A_CONTAINEDBY",
    "A_OVERLAPS",
    "CASEI",
    "ACCENTI",
    "DATE",
    "TIMESTAMP",
    "INTERVAL",
    "POINT",
    "LINESTRING",
    "POLYGON",
    "MULTIPOINT",
    "MULTILINESTRING",
    "MULTIPOLYGON",
    "GEOMETRYCOLLECTION",
    "BBOX",
].map(createFunctionCompetion);
const cqlBuiltinFunctionCompletionSource = () => {
    return ifIn(["Identifier"], completeFromList(functionCompletionList));
};
const cqlFunctionCompletionSource = (fns) => {
    return ifIn(["Identifier"], completeFromList(fns.map(createFunctionCompetion)));
};
const cqlIdentifierCompletionSource = (identifiers) => {
    return ifIn(["Identifier"], completeFromList(identifiers.map((id) => ({ label: id, type: "variable" }))));
};

function isType(arg, type, optional) {
    if (optional && !arg) {
        return true;
    }
    return (arg === null || arg === void 0 ? void 0 : arg.name) === type;
}
const cqlLinterExtension = /*@__PURE__*/linter(view => {
    let diagnostics = [];
    syntaxTree(view.state).cursor().iterate(node => {
        const child1 = node.node.firstChild;
        const child2 = child1 === null || child1 === void 0 ? void 0 : child1.nextSibling;
        const child3 = child2 === null || child2 === void 0 ? void 0 : child2.nextSibling;
        if (node.name === "TemporalPredicate") {
            if (!isType(child2, "TemporalExpression") || !isType(child3, "TemporalExpression")) {
                diagnostics.push({
                    from: node.from,
                    to: node.to,
                    severity: "error",
                    message: "Expecting two temporal expression arguments.",
                });
            }
        }
        else if (node.name === "SpatialPredicate") {
            if (!isType(child2, "GeomExpression") || !isType(child3, "GeomExpression")) {
                diagnostics.push({
                    from: node.from,
                    to: node.to,
                    severity: "error",
                    message: "Expecting two spatial expression arguments.",
                });
            }
        }
        else if (node.name === "ArrayPredicate") {
            if (!isType(child2, "ArrayOperand") || !isType(child3, "ArrayOperand")) {
                diagnostics.push({
                    from: node.from,
                    to: node.to,
                    severity: "error",
                    message: "Expecting two array expression arguments.",
                });
            }
        }
        else if (node.name === "Point") {
            if (!isType(child1, "NumericLiteral") || !isType(child2, "NumericLiteral") || !isType(child3, "NumericLiteral", true)) {
                diagnostics.push({
                    from: node.from,
                    to: node.to,
                    severity: "error",
                    message: "Expecting 2-3 numeric arguments.",
                });
            }
        }
    });
    return diagnostics;
});

function cql() {
    const extensions = [
        cqlKeywordCompletionExtension,
        cqlConstantCompletionExtension,
        cqlBuiltinFunctionCompletionExtension,
        cqlDateInstanceCompletionExtension,
        cqlTimestampInstanceCompletionExtension,
        cqlLinterExtension
    ];
    return new LanguageSupport(cqlLanguage, extensions);
}
const cqlLanguage = /*@__PURE__*/LRLanguage.define({
    parser: /*@__PURE__*/parser.configure({}),
    languageData: {
        closeBrackets: { brackets: ["(", "'", '"'] },
    },
});
const cqlKeywordCompletionExtension = /*@__PURE__*/cqlLanguage.data.of({
    autocomplete: /*@__PURE__*/cqlKeywordCompletionSource(),
});
const cqlConstantCompletionExtension = /*@__PURE__*/cqlLanguage.data.of({
    autocomplete: /*@__PURE__*/cqlConstantCompletionSource(),
});
const cqlBuiltinFunctionCompletionExtension = /*@__PURE__*/cqlLanguage.data.of({
    autocomplete: /*@__PURE__*/cqlBuiltinFunctionCompletionSource(),
});
const cqlDateInstanceCompletionExtension = /*@__PURE__*/cqlLanguage.data.of({
    autocomplete: /*@__PURE__*/cqlDateInstanceCompletionSource(),
});
const cqlTimestampInstanceCompletionExtension = /*@__PURE__*/cqlLanguage.data.of({
    autocomplete: /*@__PURE__*/cqlTimestampInstanceCompletionSource(),
});

export { cql, cqlFunctionCompletionSource, cqlIdentifierCompletionSource, cqlLanguage };
